// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable:no-any max-func-body-length no-empty no-require-imports no-var-requires
if (Reflect.metadata === undefined) {
    require('reflect-metadata');
}
const once = require("once");
const path = require("path");
const stream_1 = require("stream");
const vscode_debugadapter_1 = require("vscode-debugadapter");
const logger_1 = require("vscode-debugadapter/lib/logger");
const messages_1 = require("vscode-debugadapter/lib/messages");
const helpers_1 = require("../common/helpers");
const DebugFactory_1 = require("./DebugClients/DebugFactory");
const serviceRegistry_1 = require("./serviceRegistry");
const types_1 = require("./types");
const DEBUGGER_CONNECT_TIMEOUT = 10000;
const MIN_DEBUGGER_CONNECT_TIMEOUT = DEBUGGER_CONNECT_TIMEOUT / 2;
class PythonDebugger extends vscode_debugadapter_1.DebugSession {
    constructor(serviceContainer, isServer) {
        super(false, isServer);
        this.serviceContainer = serviceContainer;
        this.client = helpers_1.createDeferred();
    }
    shutdown(processId) {
        if (this.killDebuggerProces && processId) {
            try {
                process.kill(processId);
            }
            catch (_a) { }
        }
        if (this.debugServer) {
            this.debugServer.Stop();
            this.debugServer = undefined;
        }
        if (this.debugClient) {
            this.debugClient.Stop();
            this.debugClient = undefined;
        }
        super.shutdown();
    }
    initializeRequest(response, args) {
        const body = response.body;
        body.supportsExceptionInfoRequest = true;
        body.supportsConfigurationDoneRequest = true;
        body.supportsConditionalBreakpoints = true;
        body.supportsSetVariable = true;
        body.supportsExceptionOptions = true;
        body.exceptionBreakpointFilters = [
            {
                filter: 'raised',
                label: 'Raised Exceptions',
                default: false
            },
            {
                filter: 'uncaught',
                label: 'Uncaught Exceptions',
                default: true
            }
        ];
        if (typeof args.supportsRunInTerminalRequest === 'boolean') {
            this.supportsRunInTerminalRequest = args.supportsRunInTerminalRequest;
        }
        this.sendResponse(response);
    }
    attachRequest(response, args) {
        this.sendResponse(response);
    }
    launchRequest(response, args) {
        this.killDebuggerProces = true;
        this.startPTVSDDebugger(args)
            .then(() => this.waitForDebuggerConnection(args))
            .then(() => this.sendResponse(response))
            .catch(ex => {
            const message = this.getErrorUserFriendlyMessage(args, ex) || 'Debug Error';
            this.sendErrorResponse(response, { format: message, id: 1 }, undefined, undefined, vscode_debugadapter_1.ErrorDestination.User);
        });
    }
    startPTVSDDebugger(args) {
        return __awaiter(this, void 0, void 0, function* () {
            const launcher = DebugFactory_1.CreateLaunchDebugClient(args, this, this.supportsRunInTerminalRequest);
            this.debugServer = launcher.CreateDebugServer(undefined, this.serviceContainer);
            const serverInfo = yield this.debugServer.Start();
            return launcher.LaunchApplicationToDebug(serverInfo);
        });
    }
    waitForDebuggerConnection(args) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                let rejected = false;
                const duration = this.getConnectionTimeout(args);
                const timeout = setTimeout(() => {
                    rejected = true;
                    reject(new Error('Timeout waiting for debugger connection'));
                }, duration);
                try {
                    yield this.debugServer.client;
                    timeout.unref();
                    if (!rejected) {
                        resolve();
                    }
                }
                catch (ex) {
                    reject(ex);
                }
            }));
        });
    }
    getConnectionTimeout(args) {
        // The timeout can be overridden, but won't be documented unless we see the need for it.
        // This is just a fail safe mechanism, if the current timeout isn't enough (let study the current behaviour before exposing this setting).
        const connectionTimeout = typeof args.connectionTimeout === 'number' ? args.connectionTimeout : DEBUGGER_CONNECT_TIMEOUT;
        return Math.max(connectionTimeout, MIN_DEBUGGER_CONNECT_TIMEOUT);
    }
    getErrorUserFriendlyMessage(launchArgs, error) {
        if (!error) {
            return;
        }
        const errorMsg = typeof error === 'string' ? error : ((error.message && error.message.length > 0) ? error.message : '');
        if (helpers_1.isNotInstalledError(error)) {
            return `Failed to launch the Python Process, please validate the path '${launchArgs.pythonPath}'`;
        }
        else {
            return errorMsg;
        }
    }
}
exports.PythonDebugger = PythonDebugger;
function startDebugger() {
    return __awaiter(this, void 0, void 0, function* () {
        const serviceContainer = serviceRegistry_1.initializeIoc();
        const debugStreamProvider = serviceContainer.get(types_1.IDebugStreamProvider);
        const { input, output } = yield debugStreamProvider.getInputAndOutputStreams();
        const isServerMode = debugStreamProvider.useDebugSocketStream;
        const protocolMessageWriter = serviceContainer.get(types_1.IProtocolMessageWriter);
        // tslint:disable-next-line:no-empty
        vscode_debugadapter_1.logger.init(() => { }, path.join(__dirname, '..', '..', '..', 'experimental_debug.log'));
        const stdin = input;
        const stdout = output;
        try {
            stdin.pause();
            const handshakeDebugOutStream = new stream_1.PassThrough();
            const handshakeDebugInStream = new stream_1.PassThrough();
            const throughOutStream = new stream_1.PassThrough();
            const throughInStream = new stream_1.PassThrough();
            const inputProtocolParser = serviceContainer.get(types_1.IProtocolParser);
            inputProtocolParser.connect(throughInStream);
            const outputProtocolParser = serviceContainer.get(types_1.IProtocolParser);
            outputProtocolParser.connect(throughOutStream);
            const protocolLogger = serviceContainer.get(types_1.IProtocolLogger);
            protocolLogger.connect(throughInStream, throughOutStream);
            function enableDisableLogging(enabled) {
                if (enabled) {
                    vscode_debugadapter_1.logger.setup(logger_1.LogLevel.Verbose, true);
                    protocolLogger.setup(vscode_debugadapter_1.logger);
                }
                else {
                    protocolLogger.dispose();
                }
            }
            // Keep track of the initialize and launch requests, we'll need to re-send these to ptvsd, for bootstrapping.
            const initializeRequest = new Promise(resolve => inputProtocolParser.on('request_initialize', resolve));
            const launchRequest = new Promise(resolve => {
                inputProtocolParser.on('request_launch', (data) => {
                    const enableLogging = data.arguments.logToFile === true;
                    enableDisableLogging(enableLogging);
                    resolve(data);
                    inputProtocolParser.dispose();
                });
            });
            // Connect our intermetiate pipes.
            throughOutStream.pipe(stdout);
            handshakeDebugOutStream.pipe(throughOutStream);
            // Lets start our debugger.
            const session = new PythonDebugger(serviceContainer, isServerMode);
            session.setRunAsServer(isServerMode);
            let debuggerProcessId;
            let terminatedEventSent = false;
            let debuggerSocket;
            const dispose = once(() => __awaiter(this, void 0, void 0, function* () {
                // Wait for sometime, untill the messages are sent out (remember, we're just intercepting streams here).
                // Also its possible PTVSD might run to completion.
                yield new Promise(resolve => setTimeout(resolve, 500));
                if (debuggerSocket) {
                    throughInStream.unpipe(debuggerSocket);
                    debuggerSocket.unpipe(throughOutStream);
                }
                if (!terminatedEventSent) {
                    // Possible VS Code has closed its stream.
                    try {
                        protocolMessageWriter.write(stdout, new vscode_debugadapter_1.TerminatedEvent());
                        // Wait for this message to go out before we proceed (the process will die after this).
                        yield new Promise(resolve => setTimeout(resolve, 100));
                    }
                    catch (_a) { }
                    terminatedEventSent = true;
                }
                session.shutdown(debuggerProcessId);
            }));
            outputProtocolParser.once('event_terminated', () => {
                terminatedEventSent = true;
                dispose().catch(() => { });
            });
            // When VS Code sends a disconnect request, PTVSD replies back with a response, but its upto us to kill the process.
            // Wait for sometime, untill the messages are sent out (remember, we're just intercepting streams here).
            // Also its possible PTVSD might run to completion.
            outputProtocolParser.once('response_disconnect', () => setTimeout(dispose, 500));
            if (!isServerMode) {
                process.on('SIGTERM', dispose);
            }
            outputProtocolParser.on('response_launch', () => __awaiter(this, void 0, void 0, function* () {
                // By now we're connected to the client.
                debuggerSocket = yield session.debugServer.client;
                // We need to handle both end and error, sometimes the socket will error out without ending (if debugee is killed).
                debuggerSocket.on('end', dispose);
                debuggerSocket.on('error', dispose);
                const debugSoketProtocolParser = serviceContainer.get(types_1.IProtocolParser);
                debugSoketProtocolParser.connect(debuggerSocket);
                // Send PTVSD a bogus launch request, and wait for it to respond.
                // This needs to be done, so PTVSD can keep track of how it was launched (whether it as for attach or launch).
                protocolMessageWriter.write(debuggerSocket, yield launchRequest);
                yield new Promise(resolve => debugSoketProtocolParser.once('response_launch', resolve));
                // The PTVSD process has launched, now send the initialize request to it.
                protocolMessageWriter.write(debuggerSocket, yield initializeRequest);
                // Keep track of processid for killing it.
                debugSoketProtocolParser.once('event_process', (proc) => debuggerProcessId = proc.body.systemProcessId);
                // Wait for PTVSD to reply back with initialized event.
                debugSoketProtocolParser.once('event_initialized', (initialized) => {
                    // Get ready for PTVSD to communicate directly with VS Code.
                    throughInStream.unpipe(handshakeDebugInStream);
                    throughInStream.pipe(debuggerSocket);
                    debuggerSocket.pipe(throughOutStream);
                    // Forward the initialized event sent by PTVSD onto VSCode.
                    // This is what will cause PTVSD to start the actualy work.
                    protocolMessageWriter.write(throughOutStream, initialized);
                });
            }));
            // Start handling requests in the session instance.
            // The session (PythonDebugger class) will only perform the bootstrapping (launching of PTVSD).
            throughInStream.pipe(handshakeDebugInStream);
            stdin.pipe(throughInStream);
            session.start(handshakeDebugInStream, handshakeDebugOutStream);
            stdin.resume();
        }
        catch (ex) {
            vscode_debugadapter_1.logger.error(`Debugger crashed.${ex.message}`);
            protocolMessageWriter.write(stdout, new messages_1.Event('error', `Debugger Error: ${ex.message}`));
            protocolMessageWriter.write(stdout, new vscode_debugadapter_1.OutputEvent(ex.toString(), 'stderr'));
        }
    });
}
process.on('uncaughtException', (err) => {
    vscode_debugadapter_1.logger.error(`Uncaught Exception: ${err && err.message ? err.message : ''}`);
    vscode_debugadapter_1.logger.error(err && err.name ? err.name : '');
    vscode_debugadapter_1.logger.error(err && err.stack ? err.stack : '');
    // Catch all, incase we have string exceptions being raised.
    vscode_debugadapter_1.logger.error(err ? err.toString() : '');
    // Wait for 1 second before we die, we need to ensure errors are written to the log file.
    setTimeout(() => process.exit(-1), 1000);
});
startDebugger().catch(ex => {
    // Not necessary except for debugging and to kill linter warning about unhandled promises.
});
//# sourceMappingURL=mainV2.js.map